/*
 *  netcom.c
 *  TULIP
 *	Berkley Sockets I/O 
 *  Created by Jim on Sat Oct 12 2002.
 *  Copyright (c) 2002 Pandora Products. All rights reserved.
 *
 *		 8-Feb-2003 JS	Create duplicate socket rtn
 *		30-Apr-2003 JS	Add local address string routine
 */

#include <sys/types.h>
#include <sys/sysctl.h>
#include <netdb.h>
#include "netcom.h"

static NETCOM_CHANNEL *base = NULL;

/****************************************************************************************
*
*  NETCOM_CHANNEL *netcom_netcreate(  int port )
*
*	INPUT:	port	- Port number to use
*
*	OUTPUT:	Pointer to open channel bound to port
*			NULL if we cannot create it
*
*   NOTE: The channel if created is added to the internal list
*
*******************************************************************************************/

NETCOM_CHANNEL *netcom_netcreate( int port )
{
	NETCOM_CHANNEL *ch;
	int val;
	int	opts;
	
	// (1) Build a channel and bail if we cannot
	
	ch = (NETCOM_CHANNEL *)MEMALLOC( sizeof(NETCOM_CHANNEL) );
	if( ch == NULL )
		return( NULL );
		
	// (1) OK, build the binding endpoint
	
	ch->addr.sin_family = AF_INET;
	ch->addr.sin_addr.s_addr = htonl(INADDR_ANY);
	ch->addr.sin_port = htons(port);
	
	// (2) Create the binding socket
	
	ch->socket = socket(AF_INET,SOCK_STREAM,0);
	if( ch->socket < 0 )
	{
		MEMFREE( ch );
		return( NULL );			// Could not create socket 
    }

	opts = 1;			// Set the reuse option on
	val = setsockopt(ch->socket,
						SOL_SOCKET,
						SO_REUSEADDR,
						&opts,sizeof(int));
		
	// (3) Connect to the endpoint

	val = bind(ch->socket,
				(const struct sockaddr *)&(ch->addr),
				sizeof(struct sockaddr_in));
				
	if( val < 0 )
	{
		close( ch->socket );		// Delete unused socket
		MEMFREE( ch );
		return( NULL );
	}
	
	// (3a) Set backlog length
	
    if (listen(ch->socket, 1) < 0) 
    {
        close(ch->socket);
        MEMFREE( ch );
        return(NULL);
    }
	
	// All done, mark this and attach to list
	
	ch->valid = NETCOM_VALID;
	ch->type = NETCOM_TCP;
	ch->state = NETCOM_OPEN;
	ch->duplicate = FALSE;
	
	if( base != NULL )
	{
		ch->next = base->next;
	}
	base = ch;		// To head of list

	return( ch );
}

/****************************************************************************************
*
*  NETCOM_CHANNEL *netcom_netduplicate(  NETCOM_CHANNEL *net )
*
*	INPUT:	net	- Duplicate this one
*
*	OUTPUT:	Pointer to open channel bound to port
*			NULL if we cannot create it
*			Mark as duplicate
*
*   NOTE: The channel if created is added to the internal list
*
*******************************************************************************************/

NETCOM_CHANNEL *netcom_netduplicate( NETCOM_CHANNEL *net )
{
	NETCOM_CHANNEL *ch;
	
	// (1) Build a channel and bail if we cannot
	
	ch = (NETCOM_CHANNEL *)MEMALLOC( sizeof(NETCOM_CHANNEL) );
	if( ch == NULL )
		return( NULL );
		
	// (1) Duplicate the socket
	
	ch->socket = net->socket;
	
	// All done, mark this and attach to list
	
	ch->valid = NETCOM_VALID;
	ch->type = NETCOM_TCP;
	ch->state = NETCOM_OPEN;
	ch->duplicate = TRUE;
	
	// Cross link the buddies
	
	ch->buddy = net;
	net->buddy = ch;
	
	if( base != NULL )
	{
		ch->next = base->next;
	}
	base = ch;		// To head of list

	return( ch );
}

/****************************************************************************************
*
*  void netcom_netdestroy( NETCOM_CHANNEL *net )
*
*	INPUT:	net	- NETCOM STRUCTURE
*
*	OUTPUT:	NONE
*
*   NOTE: Close the channel
*
*******************************************************************************************/

void netcom_netdestroy( NETCOM_CHANNEL *net )
{
	NETCOM_CHANNEL *ch,*prev;
	
	// (1) Remove this from the list
	
	ch = base;
	prev = NULL;
	while( ch != NULL )
	{
		if( ch == net )
		{
			break;
		}
		else
		{
			prev = ch;
			ch = ch->next;
		}
	}
	
	if( ch == NULL )
		return;					// NOT FOUND
		
	// If found, two cases, prev == null then
	// we are first in list link around from start
	// If prev != null then use it to link around
	
	if( prev == NULL )
	{
		base = ch->next;
	}
	else
	{
		prev->next = ch->next;
	}
	
	// (2) Ok net is unlinked so destroy it
	// Close the sockets & free the memory
	
	if( !net->duplicate )
	{
		// Only come here if we are NOT a duplicate
		
		if( net->socket > 0 )
			close( net->socket );
		if( net->to_socket > 0 )
			close( net->to_socket );
	}
	MEMFREE( net );

}

/******************************************************************************************
								*** NETCOM I/O ***
*******************************************************************************************/
/****************************************************************************************
*
*  TG_BOOL netcom_select(int msec,NETCOM_CHANNEL *net)	- Check for connection attemp
*
*	INPUT:	msec	- # Seconds to wait for input on each channel
*
*	OUTPUT:	TRUE if connection attempted and accepted
*	
*******************************************************************************************/

TG_BOOL netcom_select(int msec,NETCOM_CHANNEL *net)
{
	struct timeval to;
	fd_set ready;
	int val;
	
	// (1) Set up to select
	
	FD_ZERO( &ready );
	FD_SET( net->socket,&ready );
	to.tv_sec = msec/100;
	to.tv_usec = 1000 * (msec%1000);
	
	// (2) Wait on a select

	val = select(net->socket+1,&ready,NULL,NULL,&to);
	if( val == 0 )
		return( FALSE );
	
	// (3) Ok, accept the socket
	

	if( FD_ISSET( net->socket,&ready) )
	{
		net->to_socket = accept(net->socket,(struct sockaddr *)0,(socklen_t *)0);
		if( net->to_socket > 0 )
		{
			// Mark socket as READY
			
			net->state = NETCOM_READY;
			
			// Also fill in the buddies to socket if we have one
			
			if( net->buddy != NULL )
			{
				net->buddy->to_socket = net->to_socket;
			}
			return( TRUE );
		}
		else
			return( FALSE );
	}
	else
		return( FALSE );
}

/****************************************************************************************
*
*  int netcom_gets( NETCOM_CHANNEL *net,char *buf,int len )	 - Read the channel
*
*	INPUT:	echo	- TRUE to echo input
*			net		- Read on the to socket
*			buf		- Buffer to fill
*			len		- With this many bytes
*
*	OUTPUT:	# bytes received
*
*	Read till CR\LF seen
*	
*******************************************************************************************/

int netcom_gets( TG_BOOL echo,NETCOM_CHANNEL *net,char *buf,int len )
{
	int count;
    int total = 0;
    char *bptr = buf;
    TG_BOOL flag = TRUE;
	
    // Loop till 
    // (a) CR/LF seen
    // (b) total characters seen
    
    while( flag )
    {
        count = read(net->to_socket,bptr,len);
        total += count;
        
        // Should we echo
        
        if( echo )
        {
            write(net->to_socket,bptr,count );
        }
        
        switch( bptr[count-1] )
        {
            case '\r':
            case '\n':			// (a) Quit now at end of line
                                flag = FALSE;
                                break;
                                
            default:			// (b) Are we over the total
                                bptr += count;
                                if( total >= len )
                                {
                                    flag = FALSE;
                                }
                                break;
        }
    }
        
	return( total );
}


/****************************************************************************************
*
*  int netcom_read( NETCOM_CHANNEL *net,char *buf,int len )	 - Read the channel
*
*	INPUT:	net		- Read on the to socket
*			buf		- Buffer to fill
*			len		- With this many bytes
*
*	OUTPUT:	# bytes received
*	
*******************************************************************************************/

int netcom_read( NETCOM_CHANNEL *net,char *buf,int len )
{
	int count;
	
	count = read(net->to_socket,buf,len);
	return( count );
}

/****************************************************************************************
*
*  int netcom_write( NETCOM_CHANNEL *net,char *buf,int len )	 - Writ3e the channel
*
*	INPUT:	net		- Write on the to socket
*			buf		- Buffer send
*			len		- Buffer len
*
*	OUTPUT:	# bytes sendt
*	
*******************************************************************************************/

int netcom_write( NETCOM_CHANNEL *net,char *buf,int len )
{
	int count;
	
	count = write(net->to_socket,buf,len);
	return( count );
}


