//*************************************************************************
/*
    Copyright (c) 2005. Pandora Products. All Rights Reserved.
*/
//*************************************************************************
/*
*  Module Name:         LINSERIAL		- Serial Support on a Linux system
*
*  Description:         This will list,open,setup and close serial ports
*						on Linux
*
*  Author:              Jim Schimpf
*
*  Revision History:	1 Dec 2005 Initial version
*
*		USER-API	PORT_LIST *ListPorts()
*					void ListPortsFree( PORT_LIST *l);
*					PORT_HANDLE *OpenPort( int psn,
*											PORT_LIST *e,
*											int baud,
*											int stop,
*											int parity);
*					ClosePort( PORT_HANDLE *h );
*					int StatPort( PORT_HANDLE *h );
*					int WritePort( char *data,int size,PORT_HANDLE *h );
*					int ReadPort( char *data,int size,PORT_HANDLE *h );
*/
//************************************************************************

#include "linserial.h"

// **** LOCAL FUNCTIONS ***

static int GetPortState( PORT_HANDLE *h );
static int SetPortState( int state,PORT_HANDLE *h );

#pragma mark -- PORT MGMT API --

/****************************************************************
*
*  PORT_LIST *ListPorts()	- Get a list of SERIAL RS-232 ports
*
*	INPUT:NONE
*
*	OUTPUT:	NULL
*	This is not supported in Linux now so just return NULL
*
*****************************************************************/

PORT_LIST *ListPorts(void)
{
	PORT_LIST *base = NULL;
	
	return( base );
}

/****************************************************************
*
*  void FreePorts(PORT_LIST *base ) - Free port list
*
*	INPUT:	base	- Port list
*
*	OUTPUT:	NONE
*
*****************************************************************/

void FreePorts(PORT_LIST *base )
{
	PORT_LIST *npl;
	
	// (1) Loop through ports and free as seen
	
	while( base != NULL )
	{
		// (2) Save the next for later
		
		npl = base->next;
		
		// (3) Free up the parts
		
		if( base->path != NULL )
			MEMFREE( base->path );
		if( base->name != NULL )
			MEMFREE( base->name );
			
		// (4) Free the main part and move to next
		
		MEMFREE( base );
		base = npl;
	}
}

#pragma mark -- PORT OPEN/CLOSE API --

/****************************************************************
*
*  PORT_HANDLE *OpenPort( int psn,PORT_LIST *p,
*						int baud,
*						PORT_STOP stop,
*						PORT_PARITY parity)		- Open a port
*
*	INPUT:	psn		- Psn in list (base = 1)
*			p		- PORT_LIST 
*			baud	- Baud rate desired
*			stop	- # Stop bits (1,2)
*			parity	- NONE/ODD/EVEN
*
*	OUTPUT: Open file handle
*			NULL if failed
*
*****************************************************************/

PORT_HANDLE *OpenPort( int psn,
						PORT_LIST *base,
						int baud,
						SER_PORT_STOP stop,
						SER_PORT_PARITY parity)
{
	PORT_HANDLE *h;
	int i;
	int error;
	PORT_LIST *p;
	
	// (0) Find the slot in the port list
	
	for( i=0, p = base; i<psn-1 && p != NULL ; i++ )
	{
		p = p->next;
	}
	
	if( p == NULL )
		return( NULL );
	
	// (1) Create the handle
	
	h = NEW_PORT_HANDLE;
	if( h == NULL )
	{
		return( NULL );		// FAILURE
	}
	h->fd = -1;				// Set bad descriptor
	
	// (1a) Fill the name and path fields
	
	i = strlen(p->path);
	h->path = (char *)MEMALLOC( i + 1 );
	if( h->path == NULL )
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	strcpy(h->path,p->path);
	
	i = strlen( p->name );
	h->name = (char *)MEMALLOC( i + 1 );
	if( h->name == NULL )
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	strcpy(h->name,p->name);
	
	// (2) Try to open the port
	
	h->fd = open(p->path, O_RDWR | O_NOCTTY | O_NONBLOCK);
	if( h->fd == -1 )
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	ioctl(h->fd, TIOCEXCL);	// Set for exclusive OPEN
	
	// (2a) Get current values fron port
	// Save originals and copy into run area
    if (tcgetattr(h->fd, &(h->orig)) == -1)
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	memcpy(&(h->run),&(h->orig),sizeof(struct termios));
	
    // (3) Set options to baud rate, 8 bits, parity, & size no flow
    // Init flags then set
	
	//*********** CONTROL FLAG SET
    h->run.c_cflag = 0;
	
	if( parity == SER_PARITY_NONE )
		h->run.c_cflag |= CS8;			// 8 Bit words
	else
	{
		h->run.c_cflag |= CS7;			// 7 Bit words with parity
		h->run.c_cflag |= PARENB;		// Parity enable
		if( parity == SER_PARITY_ODD )
			h->run.c_cflag |= PARODD;	// ODD parity
	}
		
	if( stop == SER_STOP_TWO )
		h->run.c_cflag |= CSTOPB;		// Two stop bits (1 by default)
		
    h->run.c_cflag |= CREAD;		// Enable input
    h->run.c_cflag |= CLOCAL;		// Ignore modem cntl lines
    
	//************* INPUT FLAGS SET
    h->run.c_iflag = 0;
    h->run.c_iflag |= IGNBRK;
	if( parity != SER_PARITY_NONE )
		h->run.c_iflag |= INPCK;	// Parity check
    
	//************* LINE FLAGS SET
    h->run.c_lflag = 0;				// No echo input
    
	//************* OPTION CHARACTERS
    for( i=0; i<NCCS; i++ )			// REF:JS12052003 Change to clear all option chars
        h->run.c_cc[i] = 0;
        
    h->run.c_cc[VMIN] = 0;			// Non blocking input
    h->run.c_cc[VTIME] = 0;			// REF:JS12052003 Set to not wait
									// for either min char or time

    printf("Options set now: CFLAG [%X] IFLAG [%X] OFLAG [%X]\n",
                    (int)h->run.c_cflag,
                    (int)h->run.c_iflag,
                    (int)h->run.c_oflag);
    // (4) Set the baud rates, if you try this in reverse order
    // then the c_flag -> 0 and will wipe out the baud rates
    // (i.e. baudrate = 0)
    
	error = cfsetospeed( &(h->run),baud );
	error = cfsetispeed( &(h->run),baud );

    // (5) Cause the new options to take effect immediately.
    if (tcsetattr(h->fd, TCSANOW, &(h->run)) == -1)
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	
    // (6) Print the new input and output baud rates & return the results
    
    printf("Input baud rate changed to %d\n", (int) cfgetispeed(&(h->run)));
    printf("Output baud rate changed to %d\n", (int) cfgetospeed(&(h->run)));

	return( h );
}

/****************************************************************
*
*  void ClosePort( PORT_HANDLE *h )	- Close the port
*
*	INPUT:	h	- Close this open port
*
*	OUTPUT: NONE
*
*****************************************************************/

void ClosePort( PORT_HANDLE *h )
{
	// (1) C heck for valid handle
	
	if( h != NULL )
	{
		// (2) If descriptor valid then reset port back to original
		
		if( h->fd != -1 )
		{
			tcsetattr(h->fd, TCSANOW, &(h->orig));
			close(h->fd);
		}
		
		// (3) Free up the allocated fields
		
		if( h->path != NULL )
			MEMFREE( h->path );
		if( h->name != NULL )
			MEMFREE( h->name );
		
		// (4) Free the handle
		
		MEMFREE( h );
	}
}
#pragma mark -- PORT I/O API --

/***********************************************************************
*
* int WritePort( char *buffer,int size,PORT_HANDLE *h) - Write to a port
*
*	INPUT:	buffer	- Buffer to send
*			size	- # Bytes to send
*			h		- To this port
*
*	OUTPUT:	# bytes written
*
************************************************************************/

int WritePort( char *buffer,int size,PORT_HANDLE *h)
{
	int rtn = -1;
	
	// (1) Good handle ?
	
	if( h != NULL )
	{
		// (2) Ok, so do I/O 
		
		rtn = write(h->fd,buffer,size );
	}
	
	return( rtn );
}

/***********************************************************************
*
* int ReadPort( char *buffer,int size,PORT_HANDLE *h) - Read from a port
*
*	INPUT:	buffer	- Buffer to read
*			size	- # Bytes to read
*			h		- From this port
*
*	OUTPUT:	# bytes read
*
************************************************************************/

int ReadPort( char *buffer, int size,PORT_HANDLE *h )
{
	int rtn = -1;
	
	// (1) Good handle ?
	
	if( h != NULL )
	{
		// (2) Ok, so do I/O 
		
		rtn = read(h->fd,buffer,size );
	}
	
	return( rtn );
}

#pragma mark -- PORT MISC --
/***********************************************************************
*
* int CTSPort( PORT_HANDLE *h ) - Return CTS state
*
*	INPUT: h	- Port handle
*
*	OUTPUT:	1 = CTS asserted, 0 - CTS not asserted
*
************************************************************************/

int CTSPort( PORT_HANDLE *h )
{
	int state = GetPortState( h );		// Get current state
	int cts;
	
	// Extract RTS state
	
	if( state & TIOCM_CTS )
	{
		cts = 1;
	}
	else
	{
		cts = 0;
	}
	
	return( cts );
}

/***********************************************************************
*
* void DTRPort( int flag,PORT_HANDLE *h) - Set DTR state
*
*	INPUT:	flag	- 1 -> Assert DTR, 0 Remove DTR
*			h		- Port to work on
*
*	OUTPUT:	NONE
*
************************************************************************/

void DTRPort( int dtr,PORT_HANDLE *h)
{
	int state = GetPortState( h );		// Get current state
	
	// (1) Set or CLEAR
	
	if( dtr )
	{
		state = state | TIOCM_DTR;		// Assert DTR
	}
	else
	{
		state = state & ~TIOCM_DTR;		// Clear DTR
	}
	
	SetPortState( state,h );
}

#pragma mark -- Support (INTERNAL) --

/***********************************************************************
*
* PORT_LIST *add_port_names(char *portstr,PORT_LIST *pl) - Add names into struct
*
*	INPUT: 	portstr		- Port path string
*			pl			- Port list struct to have names added
*						NOTE: Struct NEW so all pointers NULL
*
*	OUTPUT:	pl	- Pointer to original pl filled out if OK
*			NULL if failure
*
************************************************************************/

static PORT_LIST *add_port_names(char *portstr,PORT_LIST *pl)
{
	int len;
	char *ptr;
	
	// (1) Get the length and put in the whole path
	
	len = strlen(portstr);
	pl->path = (char *)MEMALLOC(len+1);
	if( pl->path == NULL )
	{
		MEMFREE( pl );			// FAILURE
		return( NULL );
	}
	strcpy(pl->path,portstr);
	
	// (2) Now walk back through the portstr and find the first "/"
	// from there to the end is the name of the port
	
	ptr = portstr;
	ptr += len;			// Move to the end of the string
	// Scan backwards through the string
	for( ; ptr >= portstr; ptr-- )
	{
		if( *ptr == '/' )
			break;				// We have found the left end of the name
								// Actually 1 less than that see below
	}
	
	ptr++;						// Move to start of name
								// Now build the name holder
	len = strlen(ptr);
	pl->name = (char *)MEMALLOC(len+1);
	if( pl->name == NULL )
	{
		MEMFREE(pl->path);	// FAILURE
		MEMFREE(pl);
		return( NULL );
	}
	strcpy(pl->name,ptr);
	
	// (3) All done return filled out struct
	
	return( pl );
}


/***********************************************************************
*
* int GetPortState(PORT_HANDLE *h)	- Get state value from tty
*
*	INPUT: 	xh		- Handle to open serial port
*
*	OUTPUT:	Terminal state as per bits in sys/ioctl.h
*			Return 0 if bad state
*
************************************************************************/

static int GetPortState( PORT_HANDLE *h )
{
    int state = 0;
	int val;
	
	// (2) Read state
	
	val = ioctl(h->fd,TIOCMGET,&state);
	if( val < 0 )
	{
		state = 0;		// ERROR
	}
	return( state );
}

/***********************************************************************
*
* int SetPortState(int state, PORT_HANDLE *h)	- Set state value for tty
*
*	INPUT:  state   - Set TTY state to this
*			xh		- Handle to open serial port
*
*	OUTPUT:	Return 1 if OK, 0 if fail
*
************************************************************************/

static int SetPortState( int state,PORT_HANDLE *h )
{
    int rtnval = 0;
	int val;
	
	// (1) Is it valid ?
	
	// (2) Set state
	
	val = ioctl(h->fd,TIOCMSET,&state);
	if( val >= 0 )
	{
		rtnval = 1;		// SUCCESS
	}
	return( rtnval );
}
