//*************************************************************************
/*
    Copyright (c) 2005. Pandora Products. All Rights Reserved.
*/
//*************************************************************************
/*
*  Module Name:         PCSERIAL		- Serial Support on a PC system
*
*  Description:         This will list,open,setup and close serial ports
*						on Windows
*
*  Author:              Jim Schimpf
*
*  Revision History:	21 Jan 2005 Initial version
*
*		USER-API	PORT_LIST *ListPorts()
*					void ListPortsFree( PORT_LIST *l);
*					PORT_HANDLE *OpenPort( PORT_LIST *e,
*											int baud,
*											int stop,
*											int parity);
*					ClosePort( PORT_HANDLE *h );
*					int StatPort( PORT_HANDLE *h );
*					int WritePort( char *data,int size,PORT_HANDLE *h );
*					int ReadPort( char *data,int size,PORT_HANDLE *h );
*/
//************************************************************************

#include "pcserial.h"

// **** LOCAL FUNCTIONS ***
#ifdef SQUIGGLE
static kern_return_t FindPorts(bool modem_flag,io_iterator_t *matchingServices);
static kern_return_t GetPortPath(int port,
								io_iterator_t serialPortIterator, 
								char *bsdPath, 
								CFIndex maxPathSize);
#endif

static PORT_LIST *add_port_names(char *portstr,PORT_LIST *pl);

#ifndef false
#define false	0
#define true	1
#endif

#pragma mark -- PORT MGMT API --

/****************************************************************
*
*  PORT_LIST *ListPorts()	- Get a list of SERIAL RS-232 ports
*
*	INPUT:NONE
*
*	OUTPUT:	Linked list of port data
*			NULL if none
*
*****************************************************************/

PORT_LIST *ListPorts(void)
{
	PORT_LIST *base = NULL;
	char portstr[256];
	int pcount = 0;
	PORT_LIST *pl;
	PORT_LIST *last = NULL;
	int i;
	
	// (1) Build a dummy port list with 9 ports
	
	for( i=1; i<10; i++ )
	{
		pl = NEW_PORT_LIST;
		
		sprintf(portstr,"\\\\.\\COM%d",i);
		pl = add_port_names(portstr,pl);
		if( pl == NULL )
			break;				// FAILURE
			
		// (3) Now add the element to the end of the list
		// Two cases:
		//		base == NULL	- First element just replace
		//		base != NULL	- Just add after last
		
		if( base == NULL )
		{
			base = pl;
			last = pl;
		}
		else
		{
			last->next = pl;
			last = pl;
		}
		pcount = pcount + 1;
	}
	return( base );
}

/****************************************************************
*
*  void FreePorts(PORT_LIST *base ) - Free port list
*
*	INPUT:	base	- Port list
*
*	OUTPUT:	NONE
*
*****************************************************************/

void FreePorts(PORT_LIST *base )
{
	PORT_LIST *npl;
	
	// (1) Loop through ports and free as seen
	
	while( base != NULL )
	{
		// (2) Save the next for later
		
		npl = base->next;
		
		// (3) Free up the parts
		
		if( base->path != NULL )
			MEMFREE( base->path );
		if( base->name != NULL )
			MEMFREE( base->name );
			
		// (4) Free the main part and move to next
		
		MEMFREE( base );
		base = npl;
	}
}

#pragma mark -- PORT OPEN/CLOSE API --

/****************************************************************
*
*  PORT_HANDLE *OpenPort( PORT_LIST *p,
*						int baud,
*						PORT_STOP stop,
*						PORT_PARITY parity)		- Open a port
*
*	INPUT:	p		- PORT_LIST element
*			baud	- Baud rate desired
*			stop	- # Stop bits (1,2)
*			parity	- NONE/ODD/EVEN
*
*	OUTPUT: Open file handle
*			NULL if failed
*
*****************************************************************/

PORT_HANDLE *OpenPort( PORT_LIST *p,
						int baud,
						SER_PORT_STOP stop,
						SER_PORT_PARITY parity)
{
	PORT_HANDLE *h = NULL;
	int i;
	
	// (0) Do we have a port list ?
	
	if( p == NULL )
		return( NULL );
	
	// (1) Create the handle
	
	h = NEW_PORT_HANDLE;
	if( h == NULL )
	{
		return( NULL );		// FAILURE
	}
	h->fd = INVALID_HANDLE_VALUE;	// Set bad descriptor
	
	// (1a) Fill the name and path fields
	
	i = strlen(p->path);
	h->path = (char *)MEMALLOC( i + 1 );
	if( h->path == NULL )
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	strcpy(h->path,p->path);
	
	i = strlen( p->name );
	h->name = (char *)MEMALLOC( i + 1 );
	if( h->name == NULL )
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	strcpy(h->name,p->name);
	
	// (2) Try to open the port
	
	h->fd = CreateFile( 	p->name,
			    	GENERIC_READ | GENERIC_WRITE,
				    0,		// Share mode
				    0,		// Security
				    OPEN_EXISTING,
				    FILE_ATTRIBUTE_NORMAL,
				    0);		// Template

	if( h->fd == INVALID_HANDLE_VALUE )
	{
		ClosePort( h );		// FAILURE
		return( NULL );
	}
	
	if( !GetCommState( h->fd, &(h->original) ) )
	{
		ClosePort( h );
		return( NULL );
	}

	if( !GetCommTimeouts( h->fd, &(h->timeouts) ) )
	{
		ClosePort( h );
		return( NULL );
	}

 	/* (3) Now set things up for our run ( baud etc)
           Get a good DCB to fill out
    */

	if( !GetCommState( h->fd,&(h->current) ) )  // Get current values
	{
		ClosePort( h );
		return( NULL );
	}

    // 29 March 2000 JS.
    // There is something undocumented in the DCB that controls
    // UART word size.  If you use the DCB above with no mod then
    // it will have the default word size, if it's 7 you are SOL
    // since then NOPARITY => EVEN PARITY.
    // What to do well we get the DCB filled out for the current
    // state, then run it through BuildCommDCB which fills out
    // the buad rate parity and stuff + sets the WORD SIZE (somewhere ?)
    // Then we fill in the parameters we want for our "real" use
    // and things are OK.
    // NOTE: I couldn't just use the DCB out of BuildCommDCB and
    //       fill in stuff below.  This failed the SET.  I don't
    //       know why at this point but this is a "fix" for the
    //       parity problem.
    //BuildCommDCB( "38400,n,8,1",&dset );
    // Fix II Found the "undocumented field - ByteSize"
    // Just set that....

    h->current.ByteSize = 8;          		// Set for 8 Bits
	h->current.BaudRate = baud;
	h->current.fBinary = true;				// Set to binary mode
	h->current.fOutxCtsFlow = false;		// No HDW flow control
	h->current.fOutxDsrFlow = false;		// No HDW flow control
	h->current.fDtrControl = DTR_CONTROL_DISABLE;	// DTR low when opened

	h->current.fDsrSensitivity = false;		// No input HDW control
	h->current.fOutX = false;				// No output XON/XOFF
	h->current.fInX = false;				// No input XON/XOFF
	h->current.fNull = false;				// Keep NUL bytes
	h->current.fRtsControl = RTS_CONTROL_DISABLE; // RTS low when opened
	h->current.fAbortOnError = false;		// Keep cranking
    h->current.fParity = true;       		// Set parity
	h->current.Parity = parity;				// Rather obvious
	h->current.StopBits = stop;				// Rather obvious again

	if( !SetCommState( h->fd,&(h->current) ) )
	{
		ClosePort( h );
		return( NULL );
	}

	return( h );
}

/****************************************************************
*
*  void ClosePort( PORT_HANDLE *h )	- Close the port
*
*	INPUT:	h	- Close this open port
*
*	OUTPUT: NONE
*
*****************************************************************/

void ClosePort( PORT_HANDLE *h )
{
	// (1) C heck for valid handle
	
	if( h != NULL )
	{
		// (2) If descriptor valid then reset port back to original
		
		if( h->fd != INVALID_HANDLE_VALUE )
		{
			SetCommState( h->fd, &(h->original));
			CloseHandle(h->fd);
		}
		
		// (3) Free up the allocated fields
		
		if( h->path != NULL )
			MEMFREE( h->path );
		if( h->name != NULL )
			MEMFREE( h->name );
		
		// (4) Free the handle
		
		MEMFREE( h );
	}

}
#pragma mark -- PORT I/O API --

//int StatPort( PORT_HANDLE *h );
/***********************************************************************
*
* int WritePort( char *buffer,int size,PORT_HANDLE *h) - Write to a port
*
*	INPUT:	buffer	- Buffer to send
*			size	- # Bytes to send
*			h		- To this port
*
*	OUTPUT:	# bytes written
*
************************************************************************/

+ WritePort( char *buffer,int size,PORT_HANDLE *h)
{
	int rtn = -1;
#ifdef SQUIGGLE	
	// (1) Good handle ?
	
	if( h != NULL )
	{
		// (2) Ok, so do I/O 
		
		rtn = write(h->fd,buffer,size );
	}
#endif	
	return( rtn );
}

/***********************************************************************
*
* int ReadPort( char *buffer,int size,PORT_HANDLE *h) - Read from a port
*
*	INPUT:	buffer	- Buffer to read
*			size	- # Bytes to read
*			h		- From this port
*
*	OUTPUT:	# bytes read
*
************************************************************************/

int ReadPort( char *buffer, int size,PORT_HANDLE *h )
{
	int rtn = -1;
#ifdef SQUIGGLE	
	// (1) Good handle ?
	
	if( h != NULL )
	{
		// (2) Ok, so do I/O 
		
		rtn = read(h->fd,buffer,size );
	}
#endif	
	return( rtn );
}

#ifdef SQUIGGLE
#pragma mark -- Support (INTERNAL) --
// ------- Support functions ------- READ OUT PORTS

/***********************************************************************
*
* FindPorts(io_iterator_t *i)	- Find an iterator to serial ports
*
*	INPUT: modem_flag				- TRUE if we are looking for a modem
*			matchingServices		- Fill with iterator to open ports
*
*	OUTPUT:	KERN_SUCCESS if OK
*
************************************************************************/
// Returns an iterator across all known modems. Caller is responsible for
// releasing the iterator when iteration is complete.
static kern_return_t FindPorts(bool modem_flag,io_iterator_t *matchingServices)
{
    kern_return_t		kernResult; 
    mach_port_t			masterPort;
    CFMutableDictionaryRef	classesToMatch;

/*! @function IOMasterPort
    @abstract Returns the mach port used to initiate communication with IOKit.
    @discussion Functions that don't specify an existing object require the IOKit master port to be passed. 
    @This function obtains that port.
    @param bootstrapPort Pass MACH_PORT_NULL for the default.
    @param masterPort The master port is returned.
    @result A kern_return_t error code. */

    kernResult = IOMasterPort(MACH_PORT_NULL, &masterPort);
    if (KERN_SUCCESS != kernResult)
    {
        printf("IOMasterPort returned %d\n", kernResult);
		goto exit;
    }
        
/*! @function IOServiceMatching
    @abstract Create a matching dictionary that specifies an IOService class match.
    @discussion A very common matching criteria for IOService is based on its class. IOServiceMatching will create a matching dictionary that specifies any IOService of a class, or its subclasses. The class is specified by C-string name.
    @param name The class name, as a const C-string. Class matching is successful on IOService's of this class or any subclass.
    @result The matching dictionary created, is returned on success, or zero on failure. The dictionary is commonly passed to IOServiceGetMatchingServices or IOServiceAddNotification which will consume a reference, otherwise it should be released with CFRelease by the caller. */

    // Serial devices are instances of class IOSerialBSDClient
    classesToMatch = IOServiceMatching(kIOSerialBSDServiceValue);
    if (classesToMatch == NULL)
    {
        printf("IOServiceMatching returned a NULL dictionary.\n");
    }
    else {
/*!
	@function CFDictionarySetValue
	Sets the value of the key in the dictionary.
	@param theDict The dictionary to which the value is to be set. If this
		parameter is not a valid mutable CFDictionary, the behavior is
		undefined. If the dictionary is a fixed-capacity dictionary and
		it is full before this operation, and the key does not exist in
		the dictionary, the behavior is undefined.
	@param key The key of the value to set into the dictionary. If a key 
		which matches this key is already present in the dictionary, only
		the value is changed ("add if absent, replace if present"). If
		no key matches the given key, the key-value pair is added to the
		dictionary. If added, the key is retained by the dictionary,
		using the retain callback provided
		when the dictionary was created. If the key is not of the sort
		expected by the key retain callback, the behavior is undefined.
	@param value The value to add to or replace into the dictionary. The value
		is retained by the dictionary using the retain callback provided
		when the dictionary was created, and the previous value if any is
		released. If the value is not of the sort expected by the
		retain or release callbacks, the behavior is undefined.
*/
		if( modem_flag )
		{
			// LOOK for MODEM HERE
			CFDictionarySetValue(classesToMatch,
                             CFSTR(kIOSerialBSDTypeKey),
							 CFSTR(kIOSerialBSDModemType));
		}
		else
		{
			CFDictionarySetValue(classesToMatch,
                             CFSTR(kIOSerialBSDTypeKey),
							 CFSTR(kIOSerialBSDRS232Type));
		}
		
        // Each serial device object has a property with key
        // kIOSerialBSDTypeKey and a value that is one of kIOSerialBSDAllTypes,
        // kIOSerialBSDModemType, or kIOSerialBSDRS232Type. You can experiment with the
        // matching by changing the last parameter in the above call to CFDictionarySetValue.
        
        // As shipped, this sample is only interested in modems,
        // so add this property to the CFDictionary we're matching on. 
        // This will find devices that advertise themselves as modems,
        // such as built-in and USB modems. However, this match won't find serial modems.
    }
    
    /*! @function IOServiceGetMatchingServices
        @abstract Look up registered IOService objects that match a matching dictionary.
        @discussion This is the preferred method of finding IOService objects currently registered by IOKit. IOServiceAddNotification can also supply this information and install a notification of new IOServices. The matching information used in the matching dictionary may vary depending on the class of service being looked up.
        @param masterPort The master port obtained from IOMasterPort().
        @param matching A CF dictionary containing matching information, of which one reference is consumed by this function. IOKitLib can contruct matching dictionaries for common criteria with helper functions such as IOServiceMatching, IOOpenFirmwarePathMatching.
        @param existing An iterator handle is returned on success, and should be released by the caller when the iteration is finished.
        @result A kern_return_t error code. */

    kernResult = IOServiceGetMatchingServices(masterPort, classesToMatch, matchingServices);    
    if (KERN_SUCCESS != kernResult)
    {
        //printf("IOServiceGetMatchingServices returned %d\n", kernResult);
		goto exit;
    }
        
exit:
    return kernResult;
}

/***********************************************************************
*
* GetPortPath(int port,
*				io_iterator_t serialPortIterator, 
*				char *bsdPath, 
*				CFIndex maxPathSize)			- Get Path to port
*
*	INPUT: 	port				- Port # to open (0 based)
*			serialPortIterator	- pointer to linked list of port
*			bsdPath				- Return port path here
*			maxPathSize			- Max length of bsdPath
*
*	OUTPUT:	KERN_SUCCESS if OK
*
************************************************************************/
// Given an iterator across a set of ports, return the BSD path to the first one.
// If no ports are found the path name is set to an empty string.
static kern_return_t GetPortPath(int port,
								io_iterator_t serialPortIterator, 
								char *bsdPath, 
								CFIndex maxPathSize)
{
    io_object_t		modemService;
    kern_return_t	kernResult = KERN_FAILURE;
    Boolean			modemFound = false;
    
    // Initialize the returned path
    *bsdPath = '\0';
    
    // Iterate across all modems found. In this example, use port to count the correct path
    
    while ((modemService = IOIteratorNext(serialPortIterator)) && !modemFound)
    {
    	if( port == 0 )
    	{
	        CFTypeRef	bsdPathAsCFString;

		// Get the callout device's path (/dev/cu.xxxxx). The callout device should almost always be
		// used: the dialin device (/dev/tty.xxxxx) would be used when monitoring a serial port for
		// incoming calls, e.g. a fax listener.
		
			bsdPathAsCFString = IORegistryEntryCreateCFProperty(modemService,
	                                                            CFSTR(kIOCalloutDeviceKey),
	                                                            kCFAllocatorDefault,
	                                                            0);
	        if (bsdPathAsCFString)
	        {
	            Boolean result;
	            
	            // Convert the path from a CFString to a C (NUL-terminated) string for use
		    	// with the POSIX open() call.
		    
		    	result = CFStringGetCString((const CFStringRef)bsdPathAsCFString,
	                                        (char *)bsdPath,
	                                        (long)maxPathSize, 
	                                        kCFStringEncodingASCII);
	            CFRelease(bsdPathAsCFString);
	            
	            if (result)
		    	{
	                //printf("BSD path: %s", bsdPath);
	                modemFound = true;
	                kernResult = KERN_SUCCESS;
	            }
	        }
	  	}
	  	else
	  	{
	  		port = port - 1;		// Next port
	  	}

	  	// Release the io_service_t now that we are done with it.
		
		(void) IOObjectRelease(modemService);
    }
        
    return kernResult;
}

#endif


/***********************************************************************
*
* PORT_LIST *add_port_names(char *portstr,PORT_LIST *pl) - Add names into struct
*
*	INPUT: 	portstr		- Port path string
*			pl			- Port list struct to have names added
*						NOTE: Struct NEW so all pointers NULL
*
*	OUTPUT:	pl	- Pointer to original pl filled out if OK
*			NULL if failure
*
************************************************************************/

static PORT_LIST *add_port_names(char *portstr,PORT_LIST *pl)
{
	int len;
	char *ptr;
	
	// (1) Get the length and put in the whole path
	
	len = strlen(portstr);
	pl->path = (char *)MEMALLOC(len+1);
	if( pl->path == NULL )
	{
		MEMFREE( pl );			// FAILURE
		return( NULL );
	}
	strcpy(pl->path,portstr);
	
	// (2) Just do it again for the name
	
	len = strlen(portstr);
	pl->name = (char *)MEMALLOC(len+1);
	if( pl->name == NULL )
	{
		MEMFREE( pl );			// FAILURE
		return( NULL );
	}
	strcpy(pl->name,portstr);
		
	// (3) All done return filled out struct
	
	return( pl );
}
