//*************************************************************************
/*
    Copyright (c) 2004. Tollgrade Communications, Inc. All Rights Reserved.
*/
//*************************************************************************
/*
*  Module Name:         lua_serial	- Lua network Support
*
*  Description:         This library supports TCP/IP I/O interface for the 
*						terminal server in LUA
*
*  Author:              Jim Schimpf
*
*  Revision History:	24 May 2004 Initial version
*
*		Functions:
*			net.write( data )
*			data = net.read()
*			count = net.stat()
*			net.sdrop()
*/
//************************************************************************

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"
#include "terms.h"
#include "tgtimer.h"
#include "lua_net.h"

/**********************************************************************
*
* net.swrite(data)		- Write data out the network port
*
*	INPUT:	data	- String to be written, error if not present
*
*	OUTPUT:	# Bytes written
*
**********************************************************************/

static int swrite (lua_State *L) 
{
	MY_ENV *env = L->env;
	char *ptr;
	int len;
	double dval;
	
	// (1) Get the object on the stack
	
	if( !lua_isstring( L,-1) )
	{
		return(0);			// Return NIL failure
	}
	ptr = (char *)lua_tostring(L,-1);
	len = strlen(ptr);
	
	// (2) Now write the value via the serial port
	
	len = netcom_write(env->tcp_socket,ptr,len);
	dval = (double)len;
	lua_pushnumber(L,dval);
	return(1);
}

/**********************************************************************
*
* data = net.read(n)		- Read data from the serial port
*
*	INPUT:	n	- # Bytes to read
*
*	OUTPUT:	Data buffer with n or less bytes or nil if failure
*
**********************************************************************/

static int sread (lua_State *L) 
{
	MY_ENV *env = L->env;
	const char *ptr;
	int len;
	double dval;
	
	// (1) Get the object on the stack
	
	if( !lua_isnumber( L,-1) )
	{
		return(0);			// Return NIL failure
	}
	dval = lua_tonumber(L,-1);
	len = (int)dval;
	
	// (2) Create a buffer and read the data into it
	
	ptr = MEMALLOC( len + 10 );
	if( ptr == NULL )
	{
		// FAILURE - return nil
		
		return(0);
	}
	
	// (2) Read in the serial data and stack the result
	
	len = netcom_read( env->tcp_socket,(char *)ptr,len);
	lua_pushlstring(L,ptr,len);
	MEMFREE((char *)ptr);
	return(1);
}

/**********************************************************************
*
* count = net.sstat()   - Return # characters waiting for input
*
*	INPUT:	data	- String to be written, error if not present
*
*	OUTPUT:	# Bytes written
*		STUBBED OUT - RETURNS 0 ALWAYS
*
**********************************************************************/

static int sstat (lua_State *L) 
{
	double dval = 0.0;
	
	lua_pushnumber(L,dval);
	return(1);
}

/**********************************************************************
*
* count = ser.drop()   - Drop TCP/IP connection
*
*	INPUT:	NONE
*
*	OUTPUT:	NONE
*			TCP/IP connection dropped
*
**********************************************************************/

static int drop (lua_State *L) 
{
	MY_ENV *env = L->env;

	close( env->tcp_socket->to_socket );		// Close the com socket
	env->tcp_socket->state = NETCOM_OPEN;		// Return to wait for conn
												// state
	return(0);
}

/**********************************************************************
*
* count = net.cstat()   - Network connection status
*
*	INPUT:	NONE
*
*	OUTPUT:	1   - Connected to network
*			0   - No connection
*
**********************************************************************/

static int cstat (lua_State *L) 
{
	MY_ENV *env = L->env;
	
	switch(  env->tcp_socket->state )
	{
		case NETCOM_READY:		// We are connected return
								// on stack
								lua_pushnumber(L,1.0);
								return( 1 );
								break;
								
		default:				return( 0 );		// Nothing
	}
	
	return( 0 );		// Default exit
}

/**********************************************************************
*   Library Create function
**********************************************************************/

static const luaL_reg netlib[] = {
  {"swrite",swrite},
  {"sread",sread},
  {"sstat",sstat},
  {"drop",drop},
  {"cstat",cstat},
  {NULL, NULL}
};

LUALIB_API int luaopen_netlib (lua_State *L) {
  luaL_openlib(L, LUA_NETNAME, netlib, 0);
  return 1;
}

