//*************************************************************************
/*
    Copyright (c) 2004. Tollgrade Communications, Inc. All Rights Reserved.
*/
//*************************************************************************
/*
*  Module Name:         lua_serial	- Lua Serial Support
*
*  Description:         This library supports serial I/O interface for the 
*						terminal server in LUA
*
*  Author:              Jim Schimpf
*
*  Revision History:	24 May 2004 Initial version
*
*		Functions:
*			ser.swrite( data )
*			data = ser.sread()
*			count = ser.sstat()
*			cts = ser.sgetstate()
*			ser.ssetstate(dtr)
*/
//************************************************************************

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"
#include "terms.h"
#include "tgtimer.h"
#include "lua_serial.h"

/**********************************************************************
*
* ser.swrite(data)		- Write data out the serial port
*
*	INPUT:	data	- String to be written, error if not present
*
*	OUTPUT:	# Bytes written
*
**********************************************************************/

static int swrite (lua_State *L) 
{
	MY_ENV *env = L->env;
	const char *ptr;
	int len;
	double dval;
	
	// (1) Get the object on the stack
	
	if( !lua_isstring( L,-1) )
	{
		return(0);			// Return NIL failure
	}
	ptr = lua_tostring(L,-1);
	len = strlen(ptr);
	
	// (2) Now write the value via the serial port
	
	len = WritePort((char *)ptr,len,env->ser);
	dval = (double)len;
	lua_pushnumber(L,dval);
	return(1);
}

/**********************************************************************
*
* data = ser.read(n)		- Read data from the serial port
*
*	INPUT:	n	- # Bytes to read
*
*	OUTPUT:	Data buffer with n or less bytes or nil if failure
*
**********************************************************************/

static int sread (lua_State *L) 
{
	MY_ENV *env = L->env;
	const char *ptr;
	int len;
	double dval;
	
	// (1) Get the object on the stack
	
	if( !lua_isnumber( L,-1) )
	{
		return(0);			// Return NIL failure
	}
	dval = lua_tonumber(L,-1);
	len = (int)dval;
	
	// (2) Create a buffer and read the data into it
	
	ptr = MEMALLOC( len + 10 );
	if( ptr == NULL )
	{
		// FAILURE - return nil
		
		return(0);
	}
	
	// (2) Read in the serial data and stack the result
	
	len = ReadPort((char *)ptr,len,env->ser);
	lua_pushlstring(L,ptr,len);
	MEMFREE((char *)ptr);
	return(1);
}

/**********************************************************************
*
* count = ser.sstat()   - Return # characters waiting for input
*
*	INPUT:	data	- String to be written, error if not present
*
*	OUTPUT:	# Bytes written
*		STUBBED OUT - RETURNS 0 ALWAYS
*
**********************************************************************/

static int sstat (lua_State *L) 
{
	double dval = 0.0;
	
	lua_pushnumber(L,dval);
	return(1);
}

/**********************************************************************
*
* cts = ser.getstate()		- Return serial line data
*
*	INPUT:	NONE
*
*	OUTPUT:	cts = 1 asserted, 0 non-asserted 
*
**********************************************************************/

static int sgetstate (lua_State *L) 
{
	MY_ENV *env = L->env;
	int cts;
	double dval;
	
	// (1) Get the line state
	
	cts = CTSPort(env->ser);
	
	// (2) Push result to stack
	
	dval = (double)cts;
	lua_pushnumber(L,dval);
	return(1);
}

/**********************************************************************
*
* data = ser.ssetstate(dtr)		- Set DTR to requested value
*
*	INPUT:	dtr - 1 assert DTR, 0 - remove dtr
*
*	OUTPUT:	NONE
*
**********************************************************************/

static int ssetstate (lua_State *L) 
{
	MY_ENV *env = L->env;
	int dtr;
	double dval;
	
	// (1) Get the object on the stack
	
	if( !lua_isnumber( L,-1) )
	{
		return(0);			// Return NIL failure
	}
	dval = lua_tonumber(L,-1);
	dtr = (int)dval;
	
	// (2) Set DTR to requested value
	
	DTRPort( dtr,env->ser );
	return( 0 );
}

/**********************************************************************
*   Library Create function
**********************************************************************/

static const luaL_reg serlib[] = {
  {"swrite",swrite},
  {"sread",sread},
  {"sstat",sstat},
  {"sgetstate",sgetstate},
  {"ssetstate",ssetstate},
  {NULL, NULL}
};

LUALIB_API int luaopen_serlib (lua_State *L) {
  luaL_openlib(L, LUA_SERNAME, serlib, 0);
  return 1;
}

