//*************************************************************************
/*
    Copyright (c) 2004. Tollgrade Communications, Inc. All Rights Reserved.
*/
//*************************************************************************
/*
*  Module Name:         lua_timer	- Lua Timer Support
*
*  Description:         This library supports timer functions
*
*  Author:              Jim Schimpf
*
*  Revision History:	24 May 2004 Initial version
*
*		Functions:
*			timer.topen()   - Open a timer
*			timer.tclose()  - Close a timer
*			timer.tstart()  - Reset timer
*			timer.tdone(t,et)- Test a timer
*/
//************************************************************************

#include "lua.h"
#include "lauxlib.h"
#include "lualib.h"
#include "terms.h"
#include "tgtimer.h"
#include "lua_timer.h"

/**********************************************************************
*
* t = timer.topen()		- Create a timer
*
*	INPUT:	NONE
*
*	OUTPUT:	Timer handle or nil
*
**********************************************************************/

static int topen (lua_State *L) 
{
	TG_TIMER_STRUCT *t;
	void *hdl;
	
	// (1) Create a timer struct if possible
	
	t = (TG_TIMER_STRUCT *)MEMALLOC( sizeof(TG_TIMER_STRUCT));
	if( t != NULL )
	{
		// (2) Init the timer
		
		TIMER_Start( t );
		
		// (2) Return as user handle to user
		
		hdl = (void *)t;
		lua_pushlightuserdata(L,hdl);
		return( 1 );
	}
	
	// (3) Failure, so return no handle
	
	return( 0 );
}
		
/**********************************************************************
*
* timer.tclose(t)		- Close a timer
*
*	INPUT:	h   - Handle to timer
*
*	OUTPUT:	NONE
*				Free timer data struct
*
**********************************************************************/

static int tclose (lua_State *L) 
{
	const void *hdl;
	
	// (1) Get the object on the stack
	
	if( !lua_islightuserdata( L,-1) )
	{
		return(0);			// Return NIL failure
	}
	hdl = lua_topointer(L,-1);
	if( hdl != NULL )
	{
		MEMFREE( (void *)hdl );
	}
	
	return(0);
}

/**********************************************************************
*
* timer.tstart(t)   - Reset timer back to 0
*
*	INPUT:	t   - Timer handle
*
*	OUTPUT:	NONE
*
**********************************************************************/

static int tstart (lua_State *L) 
{
	TG_TIMER_STRUCT *t;
	
	// (1) Get the object on the stack
	
	if( !lua_islightuserdata( L,-1) )
	{
		return(0);			// Return NIL failure
	}
	
	// (2) Get Timer handle and reset to 0
	
	t = (TG_TIMER_STRUCT *)lua_topointer(L,-1);
	TIMER_Start( t );
	
	return(0);
}

/**********************************************************************
*
* timer.tdone(t,et)		- Check timer for done
*
*	INPUT:	t		- Timer handle
*			et		- Timer period in ms
*
*	OUTPUT:	1   if Timer done
*			nil if not
*
**********************************************************************/

static int tdone (lua_State *L) 
{
	TG_TIMER_STRUCT *t;
	double dval;
	int et;
	
	// (1) Get the timer on the stack
	
	if( !lua_islightuserdata( L,-2) )
	{
		return(0);			// Return NIL failure
	}
	t = (TG_TIMER_STRUCT *)lua_topointer(L,-2);
	
	// (2) Get the et off the stack
	
	if( !lua_isnumber(L,-1) )
	{
		return( 0 );		// Return NIL failure
	}
	
	dval = lua_tonumber(L,-1);
	et = (int)(dval + 0.5);
	
	// (3) Now look at result
	
	if( TIMER_Finished(t,et) )
	{
		dval = 1.0;
		lua_pushnumber( L,dval );		// Return TRUE
		return( 1 );
	}
	else
	{
		// Failure - just return NIL
		
		return( 0 );
	}
	
	return(0);
}


/**********************************************************************
*   Library Create function
**********************************************************************/

static const luaL_reg timerlib[] = {
  {"topen",topen},
  {"tclose",tclose},
  {"tstart",tstart},
  {"tdone",tdone},
  {NULL, NULL}
};

LUALIB_API int luaopen_timerlib (lua_State *L) {
  luaL_openlib(L, LUA_TIMERNAME, timerlib, 0);
  return 1;
}

