//*****************************************************************************
/*
    Module Name: SMC.H

    Description:    This is the task and API for Serial Multiplexed
					Communications.
					After the task is started then other tasks can open
					the API and use the communications.
                    
                    :<Size/Socket><Data><Cksum>
                    :				- Marks packet start
                    <Size/Socket>	- Bits 0/1 = Socket #
                                      Bits 7-2 = Size of message
                    <Data>			- 1-63 bytes of data
					<Cksum>			- Mod 256 cksum over the whole packet
    Notes:

    Revision History:

       26-Jan-2003	JS	Initial Version
       28-Jan-2003	JS	Got message through
        7-Feb-2003	JS	Add READ/WRITE
        1-Jul-2003	JS	Pass baud rate in on open
*/
//*****************************************************************************

#include <string.h>
#include "smcproto.h"
#include "smc.h"

/*****************************************************************************
					*** SMC TASK ***
*****************************************************************************/
#ifdef SQUIGGLE
void smc(void)
{
    SMC_HANDLE *h1,*h2,*h3,*h4,*h5;
    SMC_PROTO *pEnv,*pEnv1;
    int ch;
    
    pEnv = SMC_proto_init(1);
    pEnv1 = SMC_proto_init(2);
    if( pEnv != NULL )
    {
        h1 = SMC_COMOPEN(pEnv,0,"COM2",0,"38400n81");
        h2 = SMC_COMOPEN(pEnv,1,"COM2",0,"38400");
        h3 = SMC_COMOPEN(pEnv,2,"COM2",0,"38400");
        h4 = SMC_COMOPEN(pEnv,3,"COM2",0,"38400");
        
        SMC_FPUTC('A',h1);
        SMC_FPUTC('B',h2);
        SMC_FPUTC('C',h3);
        SMC_FPUTC('D',h4);
    }
    
    SMC_COMCLOSE(h1);
    SMC_COMCLOSE(h2);
    SMC_COMCLOSE(h3);
    SMC_COMCLOSE(h4);
    
    if( pEnv1 != NULL )
    {
        h1 = SMC_COMOPEN(pEnv1,0,"COM2",0,"38400n81");
        h2 = SMC_COMOPEN(pEnv1,1,"COM2",0,"38400");
        h3 = SMC_COMOPEN(pEnv1,2,"COM2",0,"38400");
        h4 = SMC_COMOPEN(pEnv1,3,"COM2",0,"38400");
        
        ch = EOF;
        while( ch == EOF )
        {
            SMC_proto_drive( pEnv1 );
            ch = SMC_FGETC(h1);
        }
        
        //ch = SMC_FGETC(h2);
        //ch = SMC_FGETC(h3);
        //ch = SMC_FGETC(h4);
    }
    
    SMC_COMCLOSE(h1);
    SMC_COMCLOSE(h2);
    SMC_COMCLOSE(h3);
    SMC_COMCLOSE(h4);

}   
#endif

/*****************************************************************************
					*** SMC API ***
*****************************************************************************/
/*****************************************************************************
*
*  void *SMC_INIT_ENV( int port )
*
*	INPUT:	port	- Physical port to use
*
*	OUTPUT:	Inited env with port opened or NULL if problem
*
*****************************************************************************/

void *SMC_INIT_ENV(int port,int baud)
{
    SMC_PROTO *pEnv;
    void *ptr;
    
    pEnv = SMC_proto_init( port,baud );
    ptr = (void *)pEnv;
    return( ptr );
}

/*****************************************************************************
*
*  void *SMC_PROTO_DRIVE( void *e )
*
*	INPUT:	e		- Protocol environment
*
*	OUTPUT:	NONE
*
*****************************************************************************/

void SMC_PROTO_DRIVE(void *e)
{
    SMC_PROTO *pEnv = (SMC_PROTO *)e;
    
    SMC_proto_drive( pEnv );
}

/*****************************************************************************
*
*  void SMC_CLOSE_ENV( void *e )
*
*	INPUT:	e		- Protocol environment
*
*	OUTPUT:	NONE
*
*****************************************************************************/

void SMC_CLOSE_ENV(void *e)
{
    SMC_PROTO *pEnv = (SMC_PROTO *)e;
    
    SMC_close_env( pEnv );
}

/*****************************************************************************
*
*  SMC_HANDLE SMC_COMOPEN( void *env, char *name,int type,char *baud )
*
*	INPUT:	env		- SMC_PROTO env pointer
*			name	- Serial port name
*			type	- Serial open type
*			baud	- Baud rate and settings
*
*	OUTPUT:	Return I/O handle
*			If name == COM2 then return SMC handle
*			If not the return std MQX handle
*
*****************************************************************************/

#define SMC_ON_BIT	1		// for testing but later would use ADDRESS LINES

SMC_HANDLE *SMC_COMOPEN(void *env,
                        int socket,
                        char *name,
                        int type,
                        char *baud)
{
	SMC_HANDLE *sh;
	SMC_PROTO_HANDLE *h;
    int val;
    SMC_PROTO *pEnv;
    
    // (1) Check input env before start and bail if bad
    
    if( env == NULL )
    {
        return( NULL );
    }
    pEnv = (SMC_PROTO *)env;
    if( pEnv->valid != SMC_PROTO_VALID )
    {
        return( NULL );
    }
    
    // (2) Set MUTEX before start
    
    val = pthread_mutex_lock( &(pEnv->handle_mutex) );
	
	// (3) Build an SMC handle first
	
	sh = NEW_SMC_HANDLE;
	if(sh == NULL )
	{
        val = pthread_mutex_unlock( &(pEnv->handle_mutex) );
		return( NULL );
	}
	sh->valid = SMC_HANDLE_VALID;
    sh->env = (void *)pEnv;
	
	// (4) Determine type of open, SMC or STD ?
	
	if( SMC_ON_BIT && strcmp(name,"COM2") == 0 )
	{
		// (5) Is there an available handle ?
		
		if( socket >= SMC_MAX_HANDLES || pEnv->handles[socket].in_use )
		{
            val = pthread_mutex_unlock( &(pEnv->handle_mutex) );
			MEMFREE( h );
			return( NULL );
		}
		
		// (6) Give out a new handle and init
		
		h = &(pEnv->handles[socket]);
		h->head = 0;
		h->tail = 0;
		h->size = 0;
		h->out_size = 0;
		h->in_use = TRUE;
		
		sh->smc = (TG_HANDLE)h;
	}
	else
	{
		// (7) Do STD open if on regular port
		
		sh->std = NULL;		// Put regular comopen here....
	}
	
	// (8) Return it to the user
		
    val = pthread_mutex_unlock( &(pEnv->handle_mutex) );
	return( sh );
}

/*****************************************************************************
*
*  SMC_HANDLE SMC_COMCLOSE( char *name,int type,char *baud )
*
*	INPUT:	h	- Open SMC handle
*
*	OUTPUT:	Determine type and close handle then free
*
*****************************************************************************/

void SMC_COMCLOSE(SMC_HANDLE *h)
{
	SMC_PROTO_HANDLE *hp;
    int val;
    SMC_PROTO *pEnv;
    
    // (1) Check handle and get env
    
    if( h == NULL || h->valid != SMC_HANDLE_VALID )
    {
        return;
    }
    pEnv = (SMC_PROTO *)h->env;
    
    // (2) Now lock handle for our work
    
    val = pthread_mutex_lock( &(pEnv->handle_mutex) );
	
	// (3) Now close depending on type
	
	if( h->smc != NULL )
	{
		hp = (SMC_PROTO_HANDLE *)h->smc;
		
		hp->in_use = FALSE;
		hp->head = 0;
		hp->tail = 0;
		hp->size = 0;
	}
	else
	{
		if( h->std != NULL )
		{
			// do MQX close here
		}
	}
	
	// (4) Tidy up memory
	
	MEMFREE( h );
    val = pthread_mutex_unlock( &(pEnv->handle_mutex) );
}

/*****************************************************************************
*
*  int SMC_FGETC( SMC_HANDLE *h )
*
*	INPUT:	h	- Open SMC handle
*
*	OUTPUT:	Return EOF if not data or input char
*
*****************************************************************************/

int	SMC_FGETC( SMC_HANDLE *h )
{
	SMC_PROTO_HANDLE *hp;
	int ch;
	int count;
	TG_BYTE buf[3];
	
	// (1) Is handle OK
	
	if( h == NULL || h->valid != SMC_HANDLE_VALID )
	{
		return(EOF);
	}
	
	// (2) Depending on type do required I/O
	
	if( h->smc )
	{
		hp = (SMC_PROTO_HANDLE *)(h->smc);
		
		count = SMC_proto_read(buf,1,hp);
		if( count == 0 )
		{
			ch = EOF;
		}
		else
		{
			ch = 0xff & buf[0];
		}
	}
	else
	{
		if( h->std )
		{
			// do stdio fgetc here....
			ch = EOF;
		}
	}
	
	return( ch );
}

/*****************************************************************************
*
*  void SMC_FPUTC( int ch,SMC_HANDLE *h )
*
*	INPUT:	ch	- character to output
*			h	- Open SMC handle
*
*	OUTPUT:	NONE
*
*****************************************************************************/

void	SMC_FPUTC( int ch,SMC_HANDLE *h )
{
	SMC_PROTO_HANDLE *hp;
	int count;
	TG_BYTE buf[3];
	
	// (1) Is handle OK
	
	if( h == NULL || (h->valid != SMC_HANDLE_VALID) )
	{
		return;
	}
	
	// (2) Depending on type do required I/O
	
	if( h->smc )
	{
		hp = (SMC_PROTO_HANDLE *)(h->smc);
		
		buf[0] = ch;
		count = SMC_proto_write(buf,1,hp);
	}
	else
	{
		if( h->std != NULL )
		{
			// do stdio fgetc here....
		}
	}
}

/*****************************************************************************
*
*	int	SMC_FWRITE( char *buffer,int size,int nelem,SMC_HANDLE *h );
*
*
*	INPUT:	buffer	- Write this buffer
*			size	- Size of element in bytes
*			nelem	- # Elements to write
*			h	- Open SMC handle
*
*	OUTPUT:	# Bytes written
*
*****************************************************************************/

int	SMC_FWRITE( char *buffer,int size,int nelem,SMC_HANDLE *h )
{
	SMC_PROTO_HANDLE *hp;
	int count;
    TG_BYTE *buf;
	
	// (1) Is handle OK
	
	if( h == NULL || (h->valid != SMC_HANDLE_VALID) )
	{
		return(-1);
	}
	
	// (2) Depending on type do required I/O
	
	if( h->smc )
	{
		hp = (SMC_PROTO_HANDLE *)(h->smc);
		
		size = size * nelem;
        buf = (TG_BYTE *)&(buffer[0]);
		count = SMC_proto_write(buf,size,hp);
	}
	else
	{
		if( h->std != NULL )
		{
			// do stdio fgetc here....
		}
	}
    
    return( count );
}


/*****************************************************************************
*
*	int	SMC_FREAD( char *buffer,int size,int nelem,SMC_HANDLE *h );
*
*	INPUT:	buffer	- Return data here
*			size	- Size of data element in bytes
*			nelem	- # Elements to read
*			h		- Open SMC handle
*
*	OUTPUT:	# Characters read, if return > 0 data in buffer
*
*****************************************************************************/

int	SMC_FREAD( char *buffer,int size,int nelem,SMC_HANDLE *h )
{
	SMC_PROTO_HANDLE *hp;
	int count;
	TG_BYTE *buf;
	
	// (1) Is handle OK
	
	if( h == NULL || h->valid != SMC_HANDLE_VALID )
	{
		return(EOF);
	}
	
	// (2) Depending on type do required I/O
	
	if( h->smc )
	{
		hp = (SMC_PROTO_HANDLE *)(h->smc);
		
        size = size * nelem;
        buf = (TG_BYTE *)&(buffer[0]);
		count = SMC_proto_read(buf,size,hp);
	}
	else
	{
		if( h->std )
		{
			// do stdio fgetc here....
			count = -1;
		}
	}
	
	return( count );
}

