//*****************************************************************************
/*
    Module Name: SMCPROTO.H

    Description:    This is the protocol layer for the Serial Mux System
                    The protocol is a full duplex exchange of unacknowledged
                    packets of the form:
                    
                    :<Size/Socket><Data><Cksum>
                    :				- Marks packet start
                    <Size/Socket>	- Bits 0/1 = Socket #
                                      Bits 7-2 = Size of message
                    <Data>			- 1-63 bytes of data
					<Cksum>			- Mod 256 cksum over the whole packet
    Notes:

    Revision History:

       26-Jan-2003	JS	Initial Version
       27-Jan-2003	JS	Add handle protection mutex
       28-Jan-2003	JS	Got message through
       30-Jan-2002 	JS	Add close env call
        3-Feb-2003	JS	Add a status checker
        2-May-2003 	JS	Make sure packet status is RESET on fails
                        event when not put into handle
        1-Jul-2003	JS	Pass baud rate in on open
	   13-Nov-2003  JS  Change to  (1) Send out sync data on start
								   (2) Set out only MAX packet size -1
								       not MAX (fixes 64 byte packet problem)
*/
//*****************************************************************************

#include "smcproto.h"

// *** Support functions

static void SMC_proto_writepkt(int socket,int size,TG_BYTE *buffer,SMC_PROTO *p);
static void SMC_proto_moveinput( SMC_PROTO *p );

/*****************************************************************************
					*** SMCPROTO API ***
*****************************************************************************/
/*****************************************************************************
*
*  SMC_PROTO *SMC_proto_init( int port,int baud )
*
*	INPUT:	port	Port # to open for serial I/O
*			baud	Port baud rate
*
*	OUTPUT:	Return with inited environment
*
*****************************************************************************/

SMC_PROTO *SMC_proto_init( int port,int baud )
{
	SMC_PROTO *p;
	int i;
	int val;
	
	// (1) Create ENV filled with NULLS
	
	p = (SMC_PROTO *)MEMALLOC(sizeof(SMC_PROTO));
	if( p == NULL )
	{
		return( NULL );
	}
    p->valid = SMC_PROTO_VALID;
    
    // (2) Can we open the serial port
    
    p->port = SER_COMOPEN(port,baud);
    if( p->port <= 0 )
    {
        MEMFREE( p );
        return( NULL );
    }
	
	// (2) Init MUTEX's & pointer back to main
	
	val = pthread_mutex_init( &p->out_mutex,NULL );
	val = pthread_mutex_init( &p->handle_mutex,NULL );
	for( i=0; i< SMC_MAX_HANDLES; i++ )
	{
		val = pthread_mutex_init( &(p->handles[i].in_mutex),NULL);
		p->handles[i].p = (void *)p;					// Store env
		p->handles[i].socket = i;						// Store socket #
        p->handles[i].valid = SMC_PROTO_HANDLE_VALID;	// Store valid mark
	}
	
	// (3) Send send out some sync data to 
	// get the other end in sync
	
	SER_FPUTC('A',p->port);
	SER_FPUTC('A',p->port);
	SER_FPUTC('A',p->port);
	SER_FPUTC('A',p->port);
	SER_FPUTC('A',p->port);
	
	// (4) Return structure to user
	
	return( p );
}

/*****************************************************************************
*
*  void SMC_proto_drive( SMC_PROTO *p )
*
*	INPUT:	p		- current protocol state
*
*	OUTPUT:	NONE
*
*****************************************************************************/

void SMC_proto_drive( SMC_PROTO *p)
{
	int ch;
    TG_BYTE buffer[SMC_MAX_IBUF];
    int size;
    int psn = 0;
	
	// (1) If struct bad then bail with FALSE
	
	if( p == NULL || p->valid != SMC_PROTO_VALID )
	{
		return;
	}
	
	// (3) Get a buffer's worth of data then enter state machine
	
    size = SER_FREAD(buffer,SMC_MAX_IBUF,p->port);
    if( size <= 0 )
    {
        return;			// Nothing to do
    }
    printf("Read [%8X] [%d]\n",(int)p,size);
	p->in_count += size;				// Count size
    
    while( psn < size )
    {
        ch = buffer[psn++];
        
        // (4) Enter state machine with character and handle
        
        switch( p->state ) 
        {
            case SMC_HEADER:	if( ch == SMC_HEADER_CH )
                                {
                                    p->cksum = ch;
                                    p->state = SMC_CNTL;
                                }
                                else
                                {
                                    printf("***[%8X][%02X][%c] BAD LEAD\n",(int)p,ch,ch);
                                }
                                break;
                                
            // Get the socket #, data size and reset data pointer
            case SMC_CNTL:		p->socket = ch & SMC_SOCKET_MASK;
                                p->data_size   = 0x3f & (ch >> 2);
                                p->data_ptr = 0;
                                p->state = SMC_DATA;
                                p->cksum += ch;
                                break;
                        
            
            // Stay here till all data collected & adv cksum
            
            case SMC_DATA:		p->data[p->data_ptr++] = ch;
                                p->cksum += ch;
                                if( p->data_ptr >= p->data_size )
                                {
                                    p->state = SMC_CKSUM;
                                }
                                break;
                                
            case SMC_CKSUM:		p->cksum &= 0xff;
                                if( p->cksum == ch )
                                {
                                    // Move to user handle if open
                                    
                                    SMC_proto_moveinput( p );
                                }
                                else
                                {
                                    printf("***[%8X][%2X][%2X] Size [%d] BAD PACKET\n",
                                                        (int)p,
                                                        (int)p->cksum,
                                                        (int)ch,p->data_size);
                                    p->state = SMC_HEADER;			// Discard packet
                                }
                                break;
                            
            default:			break;					// ????
        }
    }
}

/*****************************************************************************
*
*  void SMC_close_env( SMC_PROTO *p )
*
*	INPUT:	p	- Pointer to valid env that is finished
*
*	OUTPUT:	NONE
*			Remove env structs and close port
*
*****************************************************************************/

void SMC_close_env( SMC_PROTO *p )
{
	int i;
	int val;
    
    // (1) Check for valid
    
    if( p == NULL || p->valid != SMC_PROTO_VALID )
    {
        return;			// Bail bad ENV
    }
    
    // (2) Close serial port
    
    SER_COMCLOSE(p->port);
    
    // (3) Delete MUTEX's 
    
    val = pthread_mutex_unlock( &(p->out_mutex) );
    val = pthread_mutex_destroy( &(p->out_mutex) );
	
	for( i=0; i< SMC_MAX_HANDLES; i++ )
	{
        val = pthread_mutex_unlock( &(p->handles[i].in_mutex) );
        val = pthread_mutex_destroy( &(p->handles[i].in_mutex) );
        
    }
    
	// (3) Free the struct
	
	MEMFREE( p );
}


/*****************************************************************************
*
* int SMC_proto_write(TG_BYTE *buffer,int size,SMC_PROTO_HANDLE *h)
*
*	INPUT:	buffer	- Data buffer
*			size	- # bytes to write
*			h		- Handle to write on
*
*	OUTPUT:	# bytes sent
*
*****************************************************************************/

int SMC_proto_write(TG_BYTE *buffer,int size,SMC_PROTO_HANDLE *h)
{
	SMC_PROTO *p;
	int send_size;
	int sent = 0;
	int val;
	
	// (1) Get the SMC environment and get output mutex
	
	p = (SMC_PROTO *)h->p;
	if( p->valid != SMC_PROTO_VALID )
	{
		return( 0 );
	}
	
	val = pthread_mutex_lock( &(p->out_mutex ) );
	
	// (2) Now send all the data on this handle out
	
	while( size > 0 )
	{
		// Make sure packet size is under MAX size
		// If equal still fix it
		
		if( size >= SMC_PROTO_MAX_DATA )
		{
			send_size = SMC_PROTO_MAX_DATA;
		}
		else
		{
			send_size = size;
		}
		
		SMC_proto_writepkt( h->socket,send_size,buffer,p);
		
		sent += send_size;
		size -= send_size;
	}
	
	// (2) Unlock MUTEX and Return amount sent
	
	val = pthread_mutex_unlock( &(p->out_mutex ) );
	return( sent );
}

/*****************************************************************************
*
* int SMC_proto_read(TG_BYTE *buffer,int size,SMC_PROTO_HANDLE *h)
*
*	INPUT:	buffer	- Data buffer to be filled
*			size	- # bytes to read
*			h		- Handle to read on
*
*	OUTPUT:	# bytes read
*
*****************************************************************************/

int SMC_proto_read(TG_BYTE *buffer,int size,SMC_PROTO_HANDLE *h)
{
	int val,i;
	SMC_PROTO *p;
	int ReturnVal = 0;
	
	// (1) Get the SMC environment and get handle mutex
	
	p = (SMC_PROTO *)h->p;
	if( p->valid != SMC_PROTO_VALID )
	{
		return( 0 );
	}
	
	val = pthread_mutex_lock( &(h->in_mutex ) );
	
	// (2) if there is data in the queue then get it
	
	if( h->size > 0 )
	{
		if( size > h->size )
		{
			size = h->size;
		}
		
		// (3) Read the data out into the buffer
		
		for( i=0; i<size; i++ )
		{
			buffer[i] = h->data[h->tail++];
			h->size--;
            if( h->tail >= SMC_MAX_IBUF )
            {
                h->tail = 0;
            }
		}
		
		ReturnVal = size;
	}
	
	// (3) Release the mutex and return the amount read
	
	val = pthread_mutex_unlock( &(h->in_mutex) );
	return( ReturnVal );
}

/*****************************************************************************
*
* int SMC_proto_status(SMC_PROTO_HANDLE *h)
*
*	INPUT:	h		- Handle to read on
*
*	OUTPUT:	# bytes available
*
*****************************************************************************/

int SMC_proto_status(SMC_PROTO_HANDLE *h)
{
	SMC_PROTO *p;
	
	// (1) Get the SMC environment and get handle mutex
	
	p = (SMC_PROTO *)h->p;
	if( p->valid != SMC_PROTO_VALID )
	{
		return( 0 );
	}
    
    // (2) Just return data buffer size 
    
    return( h->size );
}

/*****************************************************************************
					*** SMC INTERNAL API ***
*****************************************************************************/
/*****************************************************************************
*
*  int SMC_proto_writepkt(int socket,int size,TG_BYTE *buffer,SMC_PROTO *p);
*
*	INPUT:	socket	- Socket destination (NOTE: only 0-3 acceptible)
*			size	- # bytes to send
*			buffer	- Data buffer
*			p		- current protocol state
*
*	OUTPUT:	NONE
*			NOTE: size <= SMC_PROTO_MAX_DATA
*
*****************************************************************************/

static void SMC_proto_writepkt(int socket,int size,TG_BYTE *buffer,SMC_PROTO *p)
{
	TG_UINT32 cksum,cntl;
	int i;
	
	// (1) Send the header
	
	SER_FPUTC( SMC_HEADER_CH,p->port );
	cksum = SMC_HEADER_CH;
	
	// (2) Build and send the control byte
	
	cntl = 0xff & ((size<<2) | (SMC_SOCKET_MASK & socket));
	cksum += cntl;
	SER_FPUTC(cntl,p->port);
	
	// (3) Send the data
	
	for( i=0; i<size; i++ )
	{
		SER_FPUTC(buffer[i],p->port);
		cksum += buffer[i];
		p->out_count++;			// Increment output counter
	}
	
	// (4) Send the checksum
	
	cksum &= 0xff;
	SER_FPUTC(cksum,p->port);
	p->out_count += 3;			// O/head count
}

/*****************************************************************************
*
*  void SMC_proto_moveinput( SMC_PROTO *p )	- Move just received packet to handle
*
*	INPUT:	p		- current protocol state
*
*	OUTPUT:	Selected handle updated with input received packet
*
*****************************************************************************/

static void SMC_proto_moveinput( SMC_PROTO *p )
{
	int i,val;
	int n;
	SMC_PROTO_HANDLE *h;
	
	// (1) If handle not open then just toss the data
	
	if(!p->handles[p->socket].in_use )
	{
        printf("**[%8X]Lost data\n",(int)p);
	}
    else
    {
        // (2) Ok, get the handle MUTEX
        
        h = &(p->handles[p->socket]);
        
        val = pthread_mutex_lock( &(h->in_mutex) );
        
        // (3) See how much data will fit into the buffer
        
        val = SMC_MAX_IBUF - h->size;
        if( val >= p->data_size )
        {
            n = p->data_size;
        }
        else
        {
            printf("**[%8X]Buffer Full\n",(int)p);
            n = val;
        }
        
        // (4) Put the data into the head of the buffer
        
        for( i=0; i < n; i++ )
        {
            h->data[h->head++] = p->data[i];
            h->size++;
            if( h->head >= SMC_MAX_IBUF )
            {
                h->head = 0;	// Loop around in buffer
            }
        }
        
        // (5) Release MUTEX & return
        
        val = pthread_mutex_unlock(&(h->in_mutex));
    }
	
	// (6) Get the main buffer ready for next run
    // REF:JS02052003 Make sure to do this for all packets
    // event rejected ones
	
	p->state = SMC_HEADER;
	p->data_size = 0;
	p->data_ptr = 0;
	
}
		
