//*************************************************************************
/*
    Copyright (c) 2004. Tollgrade Communications, Inc. All Rights Reserved.
*/
//*************************************************************************
/*
*  Module Name:         TERMS	- Terminal Server
*
*  Description:         This becomes a terminal server which
*						connects a TCP/IP port to a serial 
*						port
*	NOTE: Set optimization to [-O,O1] only for DEPLOYMENT build
*
*  Author:              Jim Schimpf
*
*  Revision History:	19 May 2004 Initial version
*						23 May 2004 Add in Lua
*						25 May 2004 First running version
*						28 Nov 2005 Change here to clear stack on failed
*									function call in runluafunct()
*						28 Nov 2005 SET Optimization for Deployment to [-O,O1] only
*									if higher will not pick up characters from
*									TCP/IP or SERIAL port
*						 1 Dec 2005 Convert to std serial port form
*						 9 Dec 2005 Set TCP port cmd line flag to correct value (tcp_p)
*
*  SYNTAX:	terms [-p <serial port #/Path>] 
*				  [-baud <baud rate>] 
*				  [-tcp_p <TCP/IP Port #>] 
*				  [-c <Lua script>
*				  [-v]
*				-p	Serial port #	DEFAULT = 1
*				-baud Baud rate for serial port DEFAULT 38400
*				-tcp_p TCP Port # 	DEFAULT = 6000
*				NOTE: Selection of port # < 1024 requires ADMIN access
*				-c Lua script DEFAULT = NONE
*				-v	Show version and information*
*
*	This program will run in 4 threads
*		Thread 0	- Main thread block till others finish
*		Thread 1	- TERMS engine
*
*		Thread 2	- Telnet receive Telnet port (block waiting for access)
*		Thread 3	- TERMS Receive telnet xmit
*
*/
//************************************************************************

#include <termios.h>
#include "terms.h"
#include "tgtimer.h"
#include "lua.h"

#include "lauxlib.h"
#include "lualib.h"
#include "lua_serial.h"		// Added TG library
#include "lua_timer.h"		// Added TG library
#include "lua_net.h"		// Added TG library

// Local functions

static void show_net( int base_port );
static int flag_srch(int argc,char *argv[],char *flag,int get_value,char **rtn );
static void print_help(void);

// Lua support functions

static void openstdlibs (lua_State *l);
static TG_BOOL runluafunct( lua_State *l, char *funct );
static PORT_HANDLE *OpenSerPort(char *sport,PORT_LIST *base,int baud );

// Application threads

static void *engine( void *env );

static void *local_tx( void *env );	// TCP IN -> SER Out
static void *local_rx( void *env );	// SER IN -> TCP Out

#pragma mark -- Main --
/************************************************************************
                    *** MAIN ROUTINE ***
************************************************************************/
int main (int argc, const char * argv[]) 
{
	int val;
    MY_ENV env;
	char *ptr;
    char *sport;
    int tcp_port;
	void *rtn;
    int baud;
	pthread_t eng,tx,rx;
	lua_State *l=NULL;		// Lua state for internal operation
    PORT_LIST *base;
	
	// (0) Init and create lua
	
	printf("TERMS TCP/IP <=> SER SERVER [%s %s]\n\n",__DATE__,__TIME__);
	l = lua_open();
	l->env = (&env);		// Lock in current ENV
	env.l = l;				// Save for use in threads
	if( l == NULL )
	{
		printf("** FAILED CANNOT START LUA ***\n");
		return( -1 );
	}
	
	// Init libraries & get port list
	
	openstdlibs(l);
	base = ListPorts();
    
	// (1) Get the command line parameters
	
	if( flag_srch( argc,(char **)argv,"-v",FALSE,&ptr) )
	{
		print_help();
		lua_close(l);		// Terminate lua
		return( 0 );
	}

	if( flag_srch( argc,(char **)argv,"-c",TRUE,&ptr) )
	{
		// Load lua script and bail if failure
		
		val =lua_dofile(l,ptr);
		if( val != 0 )
		{
			print_help();
			lua_close(l);		// Terminate lua
			return( 0 );
		}
	}
    
    // (2) Get the parameters
    // Serial port & baud rate
    // # Channnels
    // TCP port
    
    if( !flag_srch(argc,(char **)argv,"-p",TRUE,&sport ) )	// Serial port #
    {
        sport = "1";			// DEFAULT = 1
    }

    // Serial port baud rate
    
    if( flag_srch(argc,(char **)argv,"-baud",TRUE,&ptr ) )	// Serial port baud rate
    {
        baud = strtoul( ptr,NULL,0 );
    }
    else
    {
        baud = B38400;		// DEFAULT = 38,400
    }
    
    if( flag_srch(argc,(char **)argv,"-tcp_p",TRUE,&ptr ) )	// # TCP/IP
    {
        tcp_port = strtoul( ptr,NULL,0 );
    }
    else
    {
        tcp_port = 6000;	// DEFAULT = 6000
    }
    
    // (2b) Show here the IP and port # used
    
	show_net(tcp_port );
	
	// (3a) Open serial port, and free port list
	// Create the ENGINE and RX/TX thread env
	
	env.ser = OpenSerPort( sport,base,baud );
	FreePorts( base );
	if( env.ser == NULL )
    {
        printf("*** FAILURE CANNOT OPEN SERIAL PORT [%s]***\n",sport);
		lua_close(l);
        return( -1 );
    }
	
	// (3b) Create the network port
	
	env.tcp_socket = netcom_netcreate( (tcp_port) );
	if( env.tcp_socket == NULL )
    {
        printf("*** FAILURE CANNOT OPEN TCP/IP PORT [%d]***\n",tcp_port);
		ClosePort( env.ser );
		lua_close(l);
        return( -1 );
    }
	
	// (4) Ready to run set up ENV & run ONSTART
	
	env.state = TERMS_START;
	env.dtr = 0;
	runluafunct(l,"ON_START");
	
	// (5) Start each of the threads
	// Start TERMS engine thread first
	// Then start a TX and RX thread for each channel
	
	val = pthread_getconcurrency();
    pthread_setconcurrency( 10 );
    val = pthread_getconcurrency();

	val = pthread_create( &(eng),NULL,&(engine),&env );		// TERMS Engine thread
	
	// Now start TX/RX 
	
	val = pthread_create(&(tx),NULL,&(local_tx),&env);
	val = pthread_create(&(rx),NULL,&(local_rx),&env);
		
	// Now just wait for engine to quit

	pthread_join( eng,&rtn );

	// Close all the used ports before exit

	ClosePort( env.ser );
	netcom_netdestroy( env.tcp_socket );
	lua_close(l);
    return 0;
}

#pragma mark -- Threads --
/***********************************************************************
                ***** TERMS THREAD ROUTINES *****
***********************************************************************/
/**********************************************************************
*
* void *engine(void *smc_env )
*
*	INPUT:	smc_env		- Pointer to SMC structure
*
*	OUTPUT:	NULL
*
*	Drive the SMC engine
*
**********************************************************************/

static void *engine( void *e )
{
	MY_ENV	*env = (MY_ENV *)e;
	TG_BOOL run = TRUE;
	TG_TIMER_STRUCT t;
	int old_cts = -1;
	static int counter = 0;
	
	// (1) Init timer for this run
	
	TIMER_Start( &t );
	
    // Now just loop forever driving the engines
    
	while( run )
	{
		// Add code here to execute periodically
		
		if( TIMER_Finished( &t,1000 ) )
		{
			counter++;					// Track seconds
			// Run on periodic state function
			
			runluafunct(env->l,"ON_PERIODIC");
			
			// Check CTS state
			env->cts = CTSPort( env->ser );
			if( env->cts != old_cts )
			{
				// Store state
				old_cts = env->cts;
				
				// Run ON_CTS function
				
				runluafunct(env->l,"ON_CTS");
				
			}
						
			TIMER_Start( &t );
		}
		
		switch( env->state )
		{
										// Just advance to connection wait
										// state
			case TERMS_START:			env->state = TERMS_CONN_WAIT;
										break;
					
										// Wait here for user connnection
			case TERMS_CONN_WAIT:		if( netcom_select( 100, env->tcp_socket ) )
										{
											if( env->tcp_socket->state == NETCOM_READY )
											{
												// Call the ON_CONN function
												
												runluafunct(env->l,"ON_CONN");
																								
												// Start the run
															
												env->state = TERMS_CONN_RUN;
											}
											else
											{
												env->state = TERMS_EXIT;	// Uh don't know...
											}
										}
										break;
			
			case TERMS_CONN_RUN:		break;		// Just do nothing, wait for state change
			
										// Just go back and wait for new one
			case TERMS_CONN_DROP:		env->state = TERMS_CONN_WAIT;
			
										// Call the ON_DROP function
										
										runluafunct(env->l,"ON_DROP");
										break;
			
			case TERMS_EXIT:			run = FALSE;
										break;
		}
	}
        
    return( NULL );
}

/**********************************************************************
*
* void *local_tx(void *s )
*
*	INPUT:	s		- Pointer to MY_ENV struct
*
*	OUTPUT:	NULL
*
*   ACTION: Wait for TCP/IP data and if present sent out serial port
*			NOTE: If TCP/IP returns < 0 then go to DROP state
*
**********************************************************************/

static void *local_tx( void *s )
{
	MY_ENV *env	= (MY_ENV *)s;
	char buffer[256];
	int count;
	TG_BOOL run = TRUE;
	
	// Loop forever till upper level drops
	
	while( run )
	{
		switch( env->state )
		{
										// Get characters (if any)
										
			case TERMS_CONN_RUN:		count = netcom_read( env->tcp_socket,
															buffer,256);
															
										// Exit on either CTS == 0 or TCP/IP drop
										
										if( count > 0 || env->cts == 0 )
										{
											// Send out the serial port
												
											count = WritePort(buffer,count,env->ser);
										}
										else
										{
											// FAILURE - Drop connection change state
												
											env->state = TERMS_CONN_DROP;
											close( env->tcp_socket->to_socket );		// Close the com socket
											env->tcp_socket->state = NETCOM_OPEN;
											
											// Drop the DTR on exit
											
											env->dtr = 0;
										}
										break;
			
			case TERMS_EXIT:			run = FALSE;		// Exit run
										break;
										
			default:					break;		// Just ignore other states
		}
	}
        
    return( NULL );
}

/**********************************************************************
*
* void *local_rx(void *s )
*
*	INPUT:	s		- Pointer to MY_ENV struct
*
*	OUTPUT:	NULL
*
*	Wait for SER data and send out TCP/IP port
*
**********************************************************************/

static void *local_rx( void *s )
{
	MY_ENV *env	= (MY_ENV *)s;
	char buffer[256];
	int count;
	TG_BOOL run = TRUE;
	
	// Loop forever till upper level drops
	
	while( run )
	{
		switch( env->state )
		{
										// Get characters (if any)
										
			case TERMS_CONN_RUN:		count = ReadPort( buffer,256,env->ser);
										if( count > 0 )
										{
											// Send out the TCP/IP port
												
											netcom_write(env->tcp_socket,buffer,count);
										}
										break;
			
			case TERMS_EXIT:			run = FALSE;		// Exit run
										break;
										
			default:					break;		// Just ignore other states
		}
	}
        
    return( NULL );
}

#pragma mark -- Support --
/***********************************************************************
				**** SUPPORT FUNCTIONS ****
***********************************************************************/

/***********************************************************************
*
*  PORT_HANDLE *OpenSerPort(char *sport,PORT_LIST *base,int baud )
*
*	INPUT:	sport	- Path or port #
*			base	- Port list (if present)
*			baud	- Open at this baud rate
*
*	OUTPUT:	Port handle to open port if created
*
**********************************************************************/

static PORT_HANDLE *OpenSerPort(char *sport,PORT_LIST *base,int baud )
{
	PORT_HANDLE *h = NULL;
	int val;
	
	// (1) If BASE == NULL then create one just for this run and use the
	// first slot
	
	if( base == NULL )
	{
		base = NEW_PORT_LIST;
		if( base != NULL )
		{
			// Ok, created use first slot
			
			base->path = sport;
			base->name = sport;
			
			h = OpenPort(1,base,baud,SER_STOP_ONE,SER_PARITY_NONE);
			
			// Tidy up
			
			MEMFREE( base );
			base = NULL;
		}
	}
	else
	{
		// (2) BASE <> NULL so convert sport to slot then open the port
		
		val = strtol(sport,NULL,0);
		h = OpenPort( val,base,baud,SER_STOP_ONE,SER_PARITY_NONE);
	}
	
	return( h );
}
			
/***********************************************************************
*
*  print_help(void)
*
*	INPUT:	NONE
*
*	OUTPUT:	NONE
*			Print program help
*
**********************************************************************/

static void print_help(void)
{
	PORT_LIST *base,*pl;
	int i;
	
	printf("TERMS - TCP -> SERIAL SERVER Version %s %s\n",__DATE__,__TIME__);
	printf("SYNTAX: terms   [-p <serial port #>]\n");
    printf("                [-baud <Baud rate>]\n");
	printf("                [-tcp_p <Port #>]\n");
	printf("                [-c <Lua connect Script>]\n");
	printf("                [-e]\n");
	printf("                [-v]\n");
	printf("     -p     <Serial port #>	DEFAULT = 1\n");
    printf("     -baud  <Rate>          DEFAULT = 38400\n");
	printf("     -tcp_p <TCP Port #>    DEFAULT = 6000\n");
	printf("     -c     <Lua script>    DEFAULT = NONE\n");
	printf("     -e Eat echo characters DEFAULT = Allow echo\n");
	printf("     -v     Show this message and return\n");
	printf("\n");
	
	// (2) Show the port list
	
	base = ListPorts();
	if( base != NULL )
	{
		printf("----- PORT LIST ----\n");
		pl = base;
		i = 1;
		while( pl != NULL )
		{
			printf("[%d] [%s] [%s]\n",i,pl->name,pl->path);
			i = i + 1;
			pl = pl->next;
		}
		
		FreePorts( base );
	}
	else
	{
		printf("**** NO SERIAL PORTS AVAILABLE ****\n");
	}
	printf("--------------------\n\n");
}

/**********************************************************************
*
* static void show_net( int channels,int base_port )
*
*	INPUT:	channels	- # Channels supported
*			base_port	- Base port number for channels
*
*	OUTPUT:	NONE
*
*	Print out IP port and ports used
*
**********************************************************************/

static void show_net( int base_port )
{
    FILE *fp;
    int i;
    char netbuf[256];
    char *ptr;
    
    // (1) Output header and start stat read
    
    printf("********* I/O Information ***********\n");
    fp = popen("/sbin/ifconfig","r");
    
    // (2) Loop through output data and 
    // show only line(s) with inet, not inet6 and not 127.0.0.1
    
    if( fp != NULL )
    {
        while( fgets(netbuf,256,fp) != NULL )
        {
            // Remove CR on EOL
            
            i = strlen( netbuf ) - 1;
            netbuf[i] = '\0';
            
            if((ptr = strstr(netbuf,"inet")) != NULL )			// Only show inet lines
            {
                if( ptr[4] != '6' )								// Skip INET6 lines
                {
                    if( strstr(netbuf,"127.0.0.1") == NULL )	// Skip loop back
                    {
                        printf("NETWORK [%s]\n",netbuf);
                    }
                }
            }
        }
        pclose(fp);
    }
    
    // (3) Show ports and channels used
    
    printf("PORTS: \n");
	printf("\tPort [%d]\n",base_port);
    printf("*************************************\n");
}

/***********************************************************************
*
*  FLAG_SRCH(argc,argv,flag,get_value,&rtn )
*
*	INPUT:	argc,argv	- Command line parameters
*		flag		- String (flag) to be found
*		get_value	- TRUE return pointer to next item
*				  FALSE no pointer returned
*		rtn		- Return pointer (NULL if not used or
*				  found)
*
*	OUTPUT:	TRUE if flag found
*		FALSE if not
*		Scan argument list for flag
*		If found then TRUE is returned and if get_value TRUE
*		return rtn pointing to string in argv[] following 
*		the flag
*		NOTE only exact matches will be found
*
**********************************************************************/

static int flag_srch(int argc,char *argv[],char *flag,int get_value,char **rtn )
{
	int i;
	char *ptr;
	
	/*  Scan through the argv's and look for a matching flag */
	
	for( i=0; i<argc; i++ )
	{
		ptr = argv[i];
		if( strcmp(ptr,flag) == 0 )
		{
		
		/*  Match found, return pointer to the following
			(if requested or NULL if not)
		*/
			i++;
			if( get_value && i < argc )
				*rtn = argv[i];
			else
				*rtn = NULL;
			return( TRUE );
		}
	}
	
	/*  Failure exit here, so return to user with FALSE */
	
	return( FALSE );
}

#pragma mark -- Lua Support --
/***********************************************************************
				**** LUA SUPPORT FUNCTIONS ****
***********************************************************************/

static const luaL_reg lualibs[] = {
  {"base", luaopen_base},
  {"table", luaopen_table},
  {"io", luaopen_io},
  {"string", luaopen_string},
  {"math", luaopen_math},
  {"debug", luaopen_debug},
  {"loadlib", luaopen_loadlib},
  /* add your libraries here */
  {"ser",luaopen_serlib},			// Added library
  {"timer",luaopen_timerlib},		// Added library
  {"net",luaopen_netlib},			// Added library
  {NULL, NULL}
};

/***********************************************************************
*
*  openstdlibs(lua_State *l)
*
*	INPUT:	l		- Open lua state
*
*	OUTPUT:	NONE init lua std libraries
*
**********************************************************************/

static void openstdlibs (lua_State *l) 
{
  const luaL_reg *lib = lualibs;
  
  for (; lib->func; lib++) 
  {
    lib->func(l);  /* open library */
    lua_settop(l, 0);  /* discard any results */
  }
}

/***********************************************************************
*
*  TG_BOOL runluafunct(lua_State *l,char *funct)
*
*	INPUT:	l		- Open lua state
*			funct   - Lua function name (no params to funct)
*
*	OUTPUT:	TRUE if function ran, FALSE if not
*
**********************************************************************/

static TG_BOOL runluafunct( lua_State *l, char *funct )
{
	TG_BOOL result = FALSE;
	int index;
	
	// (1) Find the function
	
	lua_pushstring(l,funct);
	lua_gettable(l,LUA_GLOBALSINDEX);
	
	// (2) Did we get the function ?
	if( lua_isfunction(l,-1) )
	{
		// (3) Run the function
		
		lua_call( l,0,0);
		result = TRUE;
	}
	else
	{
		// Failed, so remove the top element from the stack
		
		index = lua_gettop( l );
		lua_remove( l,index );
	}
	
	return( result );
}

